import gensim
import numpy as np
import pandas as pd
import pytest
import semantic_version
from gensim.models import KeyedVectors
from wefe.datasets import load_weat
from wefe.metrics import RND, WEAT
from wefe.query import Query
from wefe.utils import (
    calculate_ranking_correlations,
    create_ranking,
    load_test_model,
    run_queries,
)
from wefe.word_embedding_model import WordEmbeddingModel

gensim_version = semantic_version.Version.coerce(gensim.__version__)


def test_load_weat_w2v():
    test_model = load_test_model()
    assert isinstance(test_model, WordEmbeddingModel)
    assert isinstance(test_model.wv, KeyedVectors)

    NUM_OF_EMBEDDINGS_IN_TEST_MODEL = 13013
    if gensim_version.major >= 4:
        assert len(test_model.wv) == NUM_OF_EMBEDDINGS_IN_TEST_MODEL
        for word in test_model.wv.key_to_index:
            assert isinstance(test_model[word], np.ndarray)
    else:
        assert len(test_model.vocab.keys()) == NUM_OF_EMBEDDINGS_IN_TEST_MODEL
        for word in test_model.vocab:
            assert isinstance(test_model[word], np.ndarray)


@pytest.fixture
def queries_and_models():
    word_sets = load_weat()

    # Create gender queries
    gender_query_1 = Query(
        [word_sets["male_terms"], word_sets["female_terms"]],
        [word_sets["career"], word_sets["family"]],
        ["Male terms", "Female terms"],
        ["Career", "Family"],
    )
    gender_query_2 = Query(
        [word_sets["male_terms"], word_sets["female_terms"]],
        [word_sets["science"], word_sets["arts"]],
        ["Male terms", "Female terms"],
        ["Science", "Arts"],
    )
    gender_query_3 = Query(
        [word_sets["male_terms"], word_sets["female_terms"]],
        [word_sets["math"], word_sets["arts_2"]],
        ["Male terms", "Female terms"],
        ["Math", "Arts"],
    )

    # Create ethnicity queries
    test_query_1 = Query(
        [word_sets["insects"], word_sets["flowers"]],
        [word_sets["pleasant_5"], word_sets["unpleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant", "Unpleasant"],
    )

    test_query_2 = Query(
        [word_sets["weapons"], word_sets["instruments"]],
        [word_sets["pleasant_5"], word_sets["unpleasant_5"]],
        ["Instruments", "Weapons"],
        ["Pleasant", "Unpleasant"],
    )

    gender_queries = [gender_query_1, gender_query_2, gender_query_3]
    negative_test_queries = [test_query_1, test_query_2]

    test_model = load_test_model()
    dummy_model_1 = test_model.wv
    dummy_model_2 = test_model.wv
    dummy_model_3 = test_model.wv

    models = [
        WordEmbeddingModel(dummy_model_1, "dummy_model_1"),
        WordEmbeddingModel(dummy_model_2, "dummy_model_2"),
        WordEmbeddingModel(dummy_model_3, "dummy_model_3"),
    ]
    return gender_queries, negative_test_queries, models


def test_run_query_input_validation(queries_and_models):

    # -----------------------------------------------------------------
    # Input checks
    # -----------------------------------------------------------------

    # Load the inputs of the fixture
    gender_queries, _, models = queries_and_models

    with pytest.raises(
        TypeError, match="queries parameter must be a list or a numpy array*"
    ):
        run_queries(WEAT, None, None)

    with pytest.raises(
        Exception, match="queries list must have at least one query instance*"
    ):
        run_queries(WEAT, [], None)

    with pytest.raises(TypeError, match="item on index 0 must be a Query instance*"):
        run_queries(WEAT, [None, None], None)

    with pytest.raises(TypeError, match="item on index 3 must be a Query instance*"):
        run_queries(WEAT, gender_queries + [None], None)

    with pytest.raises(
        TypeError,
        match="word_embeddings_models parameter must be a list or a " "numpy array*",
    ):
        run_queries(WEAT, gender_queries, None)

    with pytest.raises(
        Exception,
        match="word_embeddings_models parameter must be a non empty list "
        "or numpy array*",
    ):
        run_queries(WEAT, gender_queries, [])

    with pytest.raises(
        TypeError, match="item on index 0 must be a WordEmbeddingModel instance*"
    ):
        run_queries(WEAT, gender_queries, [None])

    with pytest.raises(
        TypeError, match="item on index 3 must be a WordEmbeddingModel instance*"
    ):
        run_queries(WEAT, gender_queries, models + [None])

    with pytest.raises(
        TypeError,
        match="When queries_set_name parameter is provided, it must be a "
        "non-empty string*",
    ):
        run_queries(WEAT, gender_queries, models, queries_set_name=None)

    with pytest.raises(
        TypeError,
        match="When queries_set_name parameter is provided, it must be a "
        "non-empty string*",
    ):
        run_queries(WEAT, gender_queries, models, queries_set_name="")

    with pytest.raises(
        TypeError,
        match="run_experiment_params must be a dict with a params" " for the metric*",
    ):
        run_queries(WEAT, gender_queries, models, metric_params=None)

    with pytest.raises(
        Exception, match="aggregate_results parameter must be a bool value*"
    ):
        run_queries(WEAT, gender_queries, models, aggregate_results=None)

    with pytest.raises(
        Exception,
        match="aggregation_function must be one of 'sum',"
        "abs_sum', 'avg', 'abs_avg' or a callable.*",
    ):
        run_queries(WEAT, gender_queries, models, aggregation_function=None)

    with pytest.raises(
        Exception,
        match="aggregation_function must be one of 'sum',"
        "abs_sum', 'avg', 'abs_avg' or a callable.*",
    ):
        run_queries(WEAT, gender_queries, models, aggregation_function="hello")

    with pytest.raises(
        Exception, match="return_only_aggregation param must be boolean.*"
    ):
        run_queries(WEAT, gender_queries, models, return_only_aggregation=None)


def test_run_queries(queries_and_models):
    def check_results_types(results, only_negative=False):
        for row in results.values:
            for value in row:
                assert isinstance(value, np.float_)
                if only_negative:
                    assert value <= 0

    # -----------------------------------------------------------------
    # Basic run_queries execution
    # -----------------------------------------------------------------

    # Load the inputs of the fixture
    gender_queries, negative_test_queries, models = queries_and_models

    results = run_queries(metric=WEAT, queries=gender_queries, models=models,)

    assert isinstance(results, pd.DataFrame)
    assert results.shape == (3, 3)

    # Check cols
    expected_cols = [
        "Male terms and Female terms wrt Career and Family",
        "Male terms and Female terms wrt Science and Arts",
        "Male terms and Female terms wrt Math and Arts",
    ]

    for given_col, expected_col in zip(results.columns, expected_cols):
        assert given_col == expected_col

    # Check index
    expected_index = ["dummy_model_1", "dummy_model_2", "dummy_model_3"]

    for given_idx, expected_idx in zip(results.index, expected_index):
        assert given_idx, expected_idx

    # Check values
    check_results_types(results)

    results = run_queries(WEAT, negative_test_queries, models)
    check_results_types(results, only_negative=True)

    # -----------------------------------------------------------------
    # run_queries with params execution
    # -----------------------------------------------------------------

    # lost_vocabulary_threshold...
    word_sets = load_weat()
    dummy_query_1 = Query(
        [["bla", "ble", "bli"], word_sets["insects"]],
        [word_sets["pleasant_9"], word_sets["unpleasant_9"]],
    )

    results = run_queries(
        WEAT, gender_queries + [dummy_query_1], models, lost_vocabulary_threshold=0.1
    )
    assert results.shape == (3, 4)
    assert results.isnull().any().any()
    check_results_types(results)

    # metric param...
    results = run_queries(
        WEAT, gender_queries, models, metric_params={"return_effect_size": True}
    )
    assert results.shape == (3, 3)
    check_results_types(results)

    # -----------------------------------------------------------------
    # run_queries with aggregation params execution
    # -----------------------------------------------------------------

    # include_average_by_embedding..
    results = run_queries(WEAT, gender_queries, models, aggregate_results=True)
    assert results.shape == (3, 4)
    check_results_types(results)

    # avg
    results = run_queries(
        WEAT,
        negative_test_queries,
        models,
        aggregate_results=True,
        aggregation_function="avg",
    )
    assert results.shape == (3, 3)
    check_results_types(results)
    agg = results.values[:, 2]
    values = results.values[:, 0:2]
    calc_agg = np.mean(values, axis=1)
    assert np.array_equal(agg, calc_agg)

    # abs avg
    results = run_queries(
        WEAT,
        negative_test_queries,
        models,
        aggregate_results=True,
        aggregation_function="abs_avg",
    )
    assert results.shape == (3, 3)
    check_results_types(results)
    agg = results.values[:, 2]
    values = results.values[:, 0:2]
    calc_agg = np.mean(np.abs(values), axis=1)
    assert np.array_equal(agg, calc_agg)

    # sum
    results = run_queries(
        WEAT,
        negative_test_queries,
        models,
        aggregate_results=True,
        aggregation_function="sum",
    )
    assert results.shape == (3, 3)
    check_results_types(results)
    agg = results.values[:, 2]
    values = results.values[:, 0:2]
    calc_agg = np.sum(values, axis=1)
    assert np.array_equal(agg, calc_agg)

    # abs_sum
    results = run_queries(
        WEAT,
        negative_test_queries,
        models,
        aggregate_results=True,
        aggregation_function="abs_sum",
    )
    assert results.shape == (3, 3)
    check_results_types(results)
    agg = results.values[:, 2]
    values = results.values[:, 0:2]
    calc_agg = np.sum(np.abs(values), axis=1)
    assert np.array_equal(agg, calc_agg)

    # custom agg function
    results = run_queries(
        WEAT,
        negative_test_queries,
        models,
        aggregate_results=True,
        aggregation_function=lambda df: -df.abs().mean(1),
    )
    assert results.shape == (3, 3)
    check_results_types(results, only_negative=True)
    agg = results.values[:, 2]
    values = results.values[:, 0:2]
    calc_agg = -np.mean(np.abs(values), axis=1)
    assert np.array_equal(agg, calc_agg)

    # return only aggregation without query name
    results = run_queries(
        WEAT,
        gender_queries,
        models,
        aggregate_results=True,
        aggregation_function="abs_avg",
        return_only_aggregation=True,
    )
    assert results.shape == (3, 1)
    check_results_types(results)
    assert (
        results.columns[-1] == "WEAT: Unnamed queries set average of abs values score"
    )

    # return only aggregation without query name
    results = run_queries(
        WEAT,
        gender_queries,
        models,
        aggregate_results=True,
        aggregation_function="abs_avg",
        queries_set_name="Gender queries",
        return_only_aggregation=True,
    )
    assert results.shape == (3, 1)
    check_results_types(results)
    assert results.columns[-1] == "WEAT: Gender queries average of abs values score"

    # return aggregation with query name
    results = run_queries(
        WEAT,
        gender_queries,
        models,
        aggregate_results=True,
        aggregation_function="abs_avg",
        queries_set_name="Gender queries",
        return_only_aggregation=False,
    )
    assert results.shape == (3, 4)
    check_results_types(results)
    assert results.columns[-1] == "WEAT: Gender queries average of abs values score"
    # -----------------------------------------------------------------
    # run_queries with generate subqueries params execution
    # -----------------------------------------------------------------

    # with this option, the gender queries must be divided in RND template
    # (2,1). with one query replicated (arts), the remaining are only 5.
    results = run_queries(RND, gender_queries, models, generate_subqueries=True)
    assert results.shape == (3, 5)
    check_results_types(results)

    # -----------------------------------------------------------------
    # run_queries full params execution
    # -----------------------------------------------------------------
    results = run_queries(
        RND,
        gender_queries,
        models,
        queries_set_name="Gender queries",
        generate_subqueries=True,
        aggregate_results=True,
        aggregation_function="abs_avg",
        return_only_aggregation=False,
        metric_params={"distance_type": "cos"},
    )
    assert results.shape == (3, 6)
    check_results_types(results)
    assert results.columns[-1] == "RND: Gender queries average of abs values score"


def test_rank_results(queries_and_models):

    gender_queries, negative_test_queries, models = queries_and_models

    results_gender = run_queries(
        WEAT,
        gender_queries,
        models,
        queries_set_name="Gender Queries",
        aggregate_results=True,
    )
    results_negative = run_queries(
        WEAT,
        negative_test_queries,
        models,
        queries_set_name="Negative Test Queries",
        aggregate_results=True,
    )

    results_gender_rnd = run_queries(
        RND,
        gender_queries,
        models,
        queries_set_name="Gender Queries",
        generate_subqueries=True,
        aggregate_results=True,
    )

    with pytest.raises(
        TypeError,
        match="All elements of results_dataframes must be a pandas "
        "Dataframe instance*",
    ):
        create_ranking([None, results_gender, results_negative])

    with pytest.raises(
        TypeError,
        match="All elements of results_dataframes must be a pandas "
        "Dataframe instance*",
    ):
        create_ranking([results_gender, results_negative, 2])

    ranking = create_ranking([results_gender, results_negative, results_gender_rnd])

    expected_ranking = pd.DataFrame(
        {
            "WEAT: Gender Queries average of abs values score": {
                "dummy_model_1": 1.0,
                "dummy_model_2": 2.0,
                "dummy_model_3": 3.0,
            },
            "WEAT: Negative Test Queries average of abs values score": {
                "dummy_model_1": 1.0,
                "dummy_model_2": 2.0,
                "dummy_model_3": 3.0,
            },
            "RND: Gender Queries average of abs values score": {
                "dummy_model_1": 1.0,
                "dummy_model_2": 2.0,
                "dummy_model_3": 3.0,
            },
        }
    )
    assert ranking.shape == (3, 3)
    assert expected_ranking.equals(ranking)

    for row in ranking.values:
        for val in row:
            assert val <= 3 and val >= 1


def test_correlations(queries_and_models):

    gender_queries, negative_test_queries, models = queries_and_models
    results_gender = run_queries(
        WEAT,
        gender_queries,
        models,
        queries_set_name="Gender Queries",
        aggregate_results=True,
    )
    results_negative = run_queries(
        WEAT,
        negative_test_queries,
        models,
        queries_set_name="Negative Test Queries",
        aggregate_results=True,
        return_only_aggregation=True,
    )

    results_gender_rnd = run_queries(
        RND,
        gender_queries,
        models,
        queries_set_name="Gender Queries",
        generate_subqueries=True,
        aggregate_results=True,
    )

    ranking = create_ranking([results_gender, results_negative, results_gender_rnd])
    assert ranking.shape == (3, 3)

    correlations = calculate_ranking_correlations(ranking)
    assert correlations.shape == (3, 3)
    assert np.array_equal(
        correlations.columns.values,
        np.array(
            [
                "WEAT: Gender Queries average of abs values score",
                "WEAT: Negative Test Queries average of abs values score",
                "RND: Gender Queries average of abs values score",
            ]
        ),
    )
    assert np.array_equal(
        correlations.index.values,
        np.array(
            [
                "WEAT: Gender Queries average of abs values score",
                "WEAT: Negative Test Queries average of abs values score",
                "RND: Gender Queries average of abs values score",
            ]
        ),
    )




#! /usr/bin/env python
"""The Word Embeddings Fairness Evaluation Framework"""

import codecs
import os

from setuptools import find_packages, setup

# get __version__ from _version.py
ver_file = os.path.join("wefe", "_version.py")
with open(ver_file) as f:
    exec(f.read())

DISTNAME = "wefe"
DESCRIPTION = "The Word Embedding Fairness Evaluation Framework"
with codecs.open("README.rst", encoding="utf-8-sig") as f:
    LONG_DESCRIPTION = f.read()
AUTHOR = "WEFE Team"
MAINTAINER = "WEFE Team"
MAINTAINER_EMAIL = "pablo.badilla@ug.uchile.cl"
URL = "https://github.com/dccuchile/wefe"
LICENSE = "new BSD"
DOWNLOAD_URL = "https://github.com/dccuchile/wefe"
VERSION = __version__
INSTALL_REQUIRES = [
    "numpy",
    "scipy",
    "scikit-learn",
    "pandas",
    "gensim",
    "plotly",
    "six",
    "requests",
    "semantic_version",
    "tqdm",
]
CLASSIFIERS = [
    "Intended Audience :: Science/Research",
    "Intended Audience :: Developers",
    "License :: OSI Approved",
    "Programming Language :: Python",
    "Topic :: Software Development",
    "Topic :: Scientific/Engineering",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
]
EXTRAS_REQUIRE = {
    "tests": ["pytest", "pytest-cov"],
    "docs": ["sphinx", "sphinx-gallery", "sphinx_rtd_theme", "numpydoc", "matplotlib"],
}

setup(
    name=DISTNAME,
    author=AUTHOR,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    description=DESCRIPTION,
    license=LICENSE,
    url=URL,
    version=VERSION,
    download_url=DOWNLOAD_URL,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    zip_safe=False,  # the package can run out of an .egg file
    classifiers=CLASSIFIERS,
    packages=find_packages(),
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.6",
    include_package_data=True,
)



"""A Word Embedding contanier based on gensim BaseKeyedVectors."""
from typing import Callable, Dict, Sequence, Union

import gensim
import numpy as np
import semantic_version

gensim_version = semantic_version.Version.coerce(gensim.__version__)
if gensim_version.major >= 4:
    from gensim.models import KeyedVectors as BaseKeyedVectors
else:
    from gensim.models.keyedvectors import BaseKeyedVectors


EmbeddingDict = Dict[str, np.ndarray]
EmbeddingSets = Dict[str, EmbeddingDict]


class WordEmbeddingModel:
    """A wrapper for Word Embedding pre-trained models.

    It can hold gensim's KeyedVectors or gensim's api loaded models.
    It includes the name of the model and some vocab prefix if needed.
    """

    def __init__(
        self, wv: BaseKeyedVectors, name: str = None, vocab_prefix: str = None
    ):
        """Initialize the word embedding model.

        Parameters
        ----------
        wv : BaseKeyedVectors.
            An instance of word embedding loaded through gensim KeyedVector
            interface or gensim's api.
        name : str, optional
            The name of the model, by default ''.
        vocab_prefix : str, optional.
            A prefix that will be concatenated with all word in the model
            vocab, by default None.

        Raises
        ------
        TypeError
            if word_embedding is not a KeyedVectors instance.
        TypeError
            if model_name is not None and not an instance of str.
        TypeError
            if vocab_prefix is not None and not an instance of str.


        Examples
        --------
        >>> from gensim.test.utils import common_texts
        >>> from gensim.models import Word2Vec
        >>> from wefe.word_embedding_model import WordEmbeddingModel

        >>> dummy_model = Word2Vec(common_texts, window=5,
        ...                        min_count=1, workers=1).wv

        >>> model = WordEmbeddingModel(dummy_model, 'Dummy model dim=10',
        ...                            vocab_prefix='/en/')
        >>> print(model.name)
        Dummy model dim=10
        >>> print(model.vocab_prefix)
        /en/


        Attributes
        ----------
        wv : BaseKeyedVectors
            The model.
        vocab :
            The vocabulary of the model (a dict with the words that have an associated
            embedding in the model).
        model_name : str
            The name of the model.
        vocab_prefix : str
            A prefix that will be concatenated with each word of the vocab
            of the model.

        """
        # Type checking
        if not isinstance(wv, BaseKeyedVectors):
            raise TypeError(
                "wv should be an instance of gensim's BaseKeyedVectors"
                ", got {}.".format(wv)
            )

        if not isinstance(name, (str, type(None))):
            raise TypeError("name should be a string or None, got {}.".format(name))

        if not isinstance(vocab_prefix, (str, type(None))):
            raise TypeError(
                "vocab_prefix should be a string or None, got {}".format(vocab_prefix)
            )

        # Assign the attributes
        self.wv = wv

        # Obtain the vocabulary
        if gensim_version.major == 4:
            self.vocab = wv.key_to_index
        else:
            self.vocab = wv.vocab

        self.vocab_prefix = vocab_prefix
        if name is None:
            self.name = "Unnamed model"
        else:
            self.name = name

    def __eq__(self, other) -> bool:
        """Check if other is the same WordEmbeddingModel that self.

        Parameters
        ----------
        other : Any
            Some object

        Returns
        -------
        bool
            True if other is a WordEmbeddingModel that have the same model, model_name
            and vocab_prefix . False in any other case
        """
        if not isinstance(other, WordEmbeddingModel):
            return False
        if self.wv != other.wv:
            return False
        if self.name != other.name:
            return False
        if self.vocab_prefix != other.vocab_prefix:
            return False
        return True

    def __getitem__(self, key: str) -> Union[np.ndarray, None]:
        """Given a word, returns its associated embedding or none if it does not exist.

        Parameters
        ----------
        key : str
            A word

        Returns
        -------
        Union[np.ndarray, None]
            The embedding associated with the word or none if none if the word does not
            exist in the model.
        """
        if key in self.vocab:
            return self.wv[key]
        else:
            return None

    def __contains__(self, key):
        """Check if a word exists in the model's vocabulary.

        Parameters
        ----------
        key : str
            Some word.

        Returns
        -------
        bool
            True if the word exists in the model's vocabulary.
        """
        return key in self.vocab

    def normalize(self):
        """Normalize word embeddings in the model by using the L2 norm.

        Use the `init_sims` function of the gensim's `KeyedVectors` class.
        **Warning**: This operation is inplace. In other words, it replaces the
        embeddings with their L2 normalized versions.

        """
        if hasattr(self.wv, "init_sims"):
            self.wv.init_sims(replace=True)
        else:
            raise TypeError("The model does not have the init_sims method implemented.")

    def update(self, word: str, embedding: np.ndarray):
        """Update the value of an embedding of the model.

        If the method is executed with a word that is not in the vocabulary, an
        exception will be raised.

        Parameters
        ----------
        word : str
            The word whose embedding will be replaced. This word must be in the model's
            vocabulary.
        embedding : np.ndarray
            An embedding representing the word. It must have the same dimensions and
            data type as the model embeddings.

        Raises
        ------
        TypeError
            if word is not a1 string.
        TypeError
            if embedding is not an np.array.
        ValueError
            if word is not in the model's vocabulary.
        ValueError
            if the embedding is not the same size as the size of the model's embeddings.
        ValueError
            if the dtype of the embedding values is not the same as the model's
            embeddings.
        """
        if not isinstance(word, str):
            raise TypeError(
                f"word should be a string, got {word} with type {type(word)}."
            )

        if word not in self:
            raise ValueError(f"word '{word}' not in model vocab.")

        if not isinstance(embedding, np.ndarray):
            raise TypeError(
                f"'{word}' new embedding should be a np.array,"
                f" got {type(embedding)}."
            )

        embedding_size = embedding.shape[0]
        if self.wv.vector_size != embedding_size:
            raise ValueError(
                f"The size of '{word}' embedding ({embedding_size}) is different from "
                f"the size of the embeddings in the model ({self.wv.vector_size})."
            )

        if not np.issubdtype(self.wv.vectors.dtype, embedding.dtype):
            raise ValueError(
                f"embedding dtype ({embedding.dtype}) is not the same of model's dtype "
                f"({self.wv.vectors.dtype})."
            )

        if gensim_version.major >= 4:
            word_index = self.wv.key_to_index[word]
        else:
            word_index = self.wv.vocab[word].index

        self.wv.vectors[word_index] = embedding

    def batch_update(
        self, words: Sequence[str], embeddings: Union[Sequence[np.ndarray], np.ndarray],
    ):
        """Update a batch of embeddings.

        This method calls `update_embedding` method with each of the word-embedding
        pairs.
        All words must be in the vocabulary, otherwise an exception will be thrown.
        Note that both `words` and `embeddings` must have the same number of elements,
        otherwise the method will raise an exception.

        Parameters
        ----------
        words : Sequence[str]
            A sequence (list, tuple or np.array) that contains the words whose
            representations will be updated.

        embeddings : Union[Sequence[np.ndarray], np.array],
            A sequence (list or tuple) or a np.array of embeddings or an np.array that
            contains all the new embeddings. The embeddings must have the same size and
            data type as the model.

        Raises
        ------
        TypeError
            if words is not a list
        TypeError
            if embeddings is not an np.ndarray
        Exception
            if words collection has not the same size of the embedding array.
        """
        if not isinstance(words, (list, tuple, np.ndarray)):
            raise TypeError(
                "words argument should be a list, tuple or np.array of strings, "
                f"got {type(words)}."
            )

        if not isinstance(embeddings, (list, tuple, np.ndarray)):
            raise TypeError(
                "embeddings should be a list, tuple or np.array, "
                f"got: {type(embeddings)}"
            )

        if len(words) != len(embeddings):
            raise ValueError(
                "words and embeddings must have the same size, got: "
                f"len(words) = {len(words)}, len(embeddings) = {len(embeddings)}"
            )

        for idx, word in enumerate(words):
            self.update(word, embeddings[idx])



"""Mean Average Cosine Similarity (MAC) implementation."""
from typing import Any, Callable, Dict, List, Union

import numpy as np
from scipy.spatial import distance
from wefe.metrics.base_metric import BaseMetric
from wefe.preprocessing import get_embeddings_from_query
from wefe.query import Query
from wefe.word_embedding_model import WordEmbeddingModel


class MAC(BaseMetric):
    """Mean Average Cosine Similarity (MAC).

    The general steps of the test are as follows [1].

    1. Embed all target and attribute words.
    2. For each target set:

        * For each word embedding in the target set:

            * For each attribute set:

                * Calculate the cosine similarity of the target embedding and
                each attribute embedding of the set.

                * Calculate the mean of the cosines similarities and save it in a array.

    3. Average all the mean cosine similarities and return the calculated score.


    The closer the value is to 1, the less biased the query will be.

    References
    ----------
    | [1]: Thomas Manzini, Lim Yao Chong,Alan W Black, and Yulia Tsvetkov.
      Black is to Criminal as Caucasian is to Police: Detecting and Removing Multiclass
      Bias in Word Embeddings.
      In Proceedings of the 2019 Conference of the North American Chapter of the
      Association for Computational Linguistics:
      Human Language Technologies, Volume 1 (Long and Short Papers), pages 615–621,
      Minneapolis, Minnesota, June 2019. Association for Computational Linguistics.
    | [2]: https://github.com/TManzini/DebiasMulticlassWordEmbedding/blob/master/Debiasing/evalBias.py
    """

    metric_template = ("n", "n")
    metric_name = "Mean Average Cosine Similarity"
    metric_short_name = "MAC"

    def _calc_s(self, t: np.ndarray, A_j: np.ndarray) -> np.number:
        """Calculate the mean cos similarity of a target embedding and a attribute set.

        Parameters
        ----------
        t : np.ndarray
            A target embedding
        A_j : np.ndarray
            An attribute embedding set.

        Returns
        -------
        float
            The mean cosine similarity between the target embedding and the attribute
            set calculated.
        """
        return np.mean([distance.cosine(t, a_i) for a_i in A_j])

    def _calc_mac(self, T, A):

        # dict that will store the s scores by target word and attribute set.
        targets_eval = {}
        # list that will store the s scores
        targets_eval_scores = []

        # T_i -> Current target set
        # t_i -> Current target embedding
        # A_j -> Current attribute set

        # for each target set
        for T_i_name, T_i_vecs in T.items():
            targets_eval[T_i_name] = {}
            # for each embedding in the current target set
            for t_i_word, t_i_vec in T_i_vecs.items():
                targets_eval[T_i_name][t_i_word] = {}
                # for each attribute set
                for A_j_name, A_j_vecs in A.items():
                    # calculate s score
                    score = self._calc_s(t_i_vec, A_j_vecs.values())
                    # add the score to the variables that will store it
                    targets_eval[T_i_name][t_i_word][A_j_name] = score
                    targets_eval_scores.append(score)

        # obtain mac by calculating the mean over the s scores in targets_eval_scores.
        mac_score = np.mean(np.array(targets_eval_scores))

        return mac_score, targets_eval

    def run_query(
        self,
        query: Query,
        model: WordEmbeddingModel,
        lost_vocabulary_threshold: float = 0.2,
        preprocessors: List[Dict[str, Union[str, bool, Callable]]] = [{}],
        strategy: str = "first",
        normalize: bool = False,
        warn_not_found_words: bool = False,
        *args: Any,
        **kwargs: Any
    ) -> Dict[str, Any]:
        """Calculate the MAC metric over the provided parameters.

        Parameters
        ----------
        query : Query
            A Query object that contains the target and attribute word sets
            for be tested.

        model : WordEmbeddingModel
            A word embedding model.

        preprocessors : List[Dict[str, Union[str, bool, Callable]]]
            A list with preprocessor options.

            A ``preprocessor`` is a dictionary that specifies what processing(s) are
            performed on each word before it is looked up in the model vocabulary.
            For example, the ``preprocessor``
            ``{'lowecase': True, 'strip_accents': True}`` allows you to lowercase
            and remove the accent from each word before searching for them in the
            model vocabulary. Note that an empty dictionary ``{}`` indicates that no
            preprocessing is done.

            The possible options for a preprocessor are:

            *   ``lowercase``: ``bool``. Indicates that the words are transformed to
                lowercase.
            *   ``uppercase``: ``bool``. Indicates that the words are transformed to
                uppercase.
            *   ``titlecase``: ``bool``. Indicates that the words are transformed to
                titlecase.
            *   ``strip_accents``: ``bool``, ``{'ascii', 'unicode'}``: Specifies that
                the accents of the words are eliminated. The stripping type can be
                specified. True uses ‘unicode’ by default.
            *   ``preprocessor``: ``Callable``. It receives a function that operates
                on each word. In the case of specifying a function, it overrides the
                default preprocessor (i.e., the previous options stop working).

            A list of preprocessor options allows you to search for several
            variants of the words into the model. For example, the preprocessors
            ``[{}, {"lowercase": True, "strip_accents": True}]``
            ``{}`` allows first to search for the original words in the vocabulary of
            the model. In case some of them are not found,
            ``{"lowercase": True, "strip_accents": True}`` is executed on these words
            and then they are searched in the model vocabulary.

        strategy : str, optional
            The strategy indicates how it will use the preprocessed words: 'first' will
            include only the first transformed word found. all' will include all
            transformed words found, by default "first".

        normalize : bool, optional
            True indicates that embeddings will be normalized, by default False

        warn_not_found_words : bool, optional
            Specifies if the function will warn (in the logger)
            the words that were not found in the model's vocabulary, by default False.

        Returns
        -------
        Dict[str, Any]
            A dictionary with the query name, the resulting score of the metric,
            and a dictionary with the distances of each attribute word
            with respect to the target sets means.

        Examples
        --------
        >>> from wefe.metrics import MAC
        >>> from wefe.query import Query
        >>> from wefe.utils import load_test_model
        >>>
        >>> query = Query(
        ...     target_sets=[
        ...         ["female", "woman", "girl", "sister", "she", "her", "hers",
        ...          "daughter"],
        ...         ["male", "man", "boy", "brother", "he", "him", "his", "son"],
        ...     ],
        ...     attribute_sets=[
        ...         ["home", "parents", "children", "family", "cousins", "marriage",
        ...          "wedding", "relatives",
        ...         ],
        ...         ["executive", "management", "professional", "corporation", "salary",
        ...          "office", "business", "career",
        ...         ],
        ...     ],
        ...     target_sets_names=["Female terms", "Male Terms"],
        ...     attribute_sets_names=["Family", "Careers"],
        ... )
        >>>
        >>> # load the model (in this case, the test model included in wefe)
        >>> model = load_test_model()
        >>>
        >>> # instance the metric and run the query
        >>> MAC().run_query(query, model) # doctest: +SKIP
        {'query_name': 'Female terms and Male Terms wrt Family and Careers',
        'result': 0.8416415235615204,
        'mac': 0.8416415235615204,
        'targets_eval': {'Female terms': {'female': {'Family': 0.9185737599618733,
            'Careers': 0.916069650076679},
            'woman': {'Family': 0.752434104681015, 'Careers': 0.9377805145923048},
            'girl': {'Family': 0.707457959651947, 'Careers': 0.9867974997032434},
            'sister': {'Family': 0.5973392464220524, 'Careers': 0.9482253392925486},
            'she': {'Family': 0.7872791914269328, 'Careers': 0.9161583095556125},
            'her': {'Family': 0.7883057091385126, 'Careers': 0.9237247597193345},
            'hers': {'Family': 0.7385367527604103, 'Careers': 0.9480051446007565},
            'daughter': {'Family': 0.5472579970955849, 'Careers': 0.9277344475267455}},
        'Male Terms': {'male': {'Family': 0.8735092766582966,
            'Careers': 0.9468009045813233},
            'man': {'Family': 0.8249392118304968, 'Careers': 0.9350165261421353},
            'boy': {'Family': 0.7106057899072766, 'Careers': 0.9879048476286698},
            'brother': {'Family': 0.6280269809067249, 'Careers': 0.9477180293761194},
            'he': {'Family': 0.8693044614046812, 'Careers': 0.8771287016716087},
            'him': {'Family': 0.8230192996561527, 'Careers': 0.888683641096577},
            'his': {'Family': 0.8876195731572807, 'Careers': 0.8920885202242061},
            'son': {'Family': 0.5764635019004345, 'Careers': 0.9220191016211174}}}}
        """
        # check the types of the provided arguments (only the defaults).
        self._check_input(query, model, locals())

        # transform query word sets into embeddings
        embeddings = get_embeddings_from_query(
            model=model,
            query=query,
            lost_vocabulary_threshold=lost_vocabulary_threshold,
            preprocessors=preprocessors,
            strategy=strategy,
            normalize=normalize,
            warn_not_found_words=warn_not_found_words,
        )

        # if there is any/some set has less words than the allowed limit,
        # return the default value (nan)
        if embeddings is None:
            return {
                "query_name": query.query_name,
                "result": np.nan,
                "mac": np.nan,
                "targets_eval": None,
            }

        # get the targets and attribute sets transformed into embeddings.
        target_sets, attribute_sets = embeddings

        mac, targets_eval = self._calc_mac(target_sets, attribute_sets)

        return {
            "query_name": query.query_name,
            "result": mac,
            "mac": mac,
            "targets_eval": targets_eval,
        }



"""Metrics Testing"""
import numpy as np
import pytest
from gensim.models.keyedvectors import KeyedVectors
from wefe.datasets.datasets import load_weat
from wefe.metrics import ECT, MAC, RIPA, RND, RNSB, WEAT
from wefe.query import Query
from wefe.word_embedding_model import WordEmbeddingModel


@pytest.fixture
def model() -> WordEmbeddingModel:
    """Load a subset of Word2vec as a testing model.

    Returns
    -------
    WordEmbeddingModel
        The loaded testing model.
    """
    w2v = KeyedVectors.load("./wefe/tests/w2v_test.kv")
    return WordEmbeddingModel(w2v, "word2vec")


@pytest.fixture
def weat_wordsets():
    return load_weat()


def test_WEAT(model, weat_wordsets):

    weat = WEAT()
    query = Query(
        [weat_wordsets["flowers"], weat_wordsets["insects"]],
        [weat_wordsets["pleasant_5"], weat_wordsets["unpleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant", "Unpleasant"],
    )
    results = weat.run_query(query, model)

    assert results["query_name"] == "Flowers and Insects wrt Pleasant and Unpleasant"
    assert isinstance(results["result"], np.number)
    assert isinstance(results["weat"], np.number)
    assert isinstance(results["effect_size"], np.number)
    assert results["result"] == results["weat"]
    assert np.isnan(results["p_value"])

    results = weat.run_query(query, model, return_effect_size=True)
    assert isinstance(results["result"], np.number)
    assert isinstance(results["weat"], np.number)
    assert isinstance(results["effect_size"], np.number)
    assert results["result"] == results["effect_size"]
    assert np.isnan(results["p_value"])

    results = weat.run_query(
        query,
        model,
        calculate_p_value=True,
        p_value_iterations=100,
        p_value_test_type="left-sided",
    )

    assert isinstance(results["result"], np.number)
    assert isinstance(results["weat"], np.number)
    assert isinstance(results["effect_size"], np.number)
    assert isinstance(results["p_value"], (float, np.number))

    results = weat.run_query(
        query,
        model,
        calculate_p_value=True,
        p_value_iterations=100,
        p_value_test_type="right-sided",
    )

    assert isinstance(results["result"], np.number)
    assert isinstance(results["weat"], np.number)
    assert isinstance(results["effect_size"], np.number)
    assert isinstance(results["p_value"], (float, np.number))

    results = weat.run_query(
        query,
        model,
        calculate_p_value=True,
        p_value_iterations=100,
        p_value_test_type="two-sided",
    )

    assert isinstance(results["result"], np.number)
    assert isinstance(results["weat"], np.number)
    assert isinstance(results["effect_size"], np.number)
    assert isinstance(results["p_value"], (float, np.number))


def test_RND(model, weat_wordsets):

    rnd = RND()
    query = Query(
        [weat_wordsets["flowers"], weat_wordsets["insects"]],
        [weat_wordsets["pleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant"],
    )
    # test with euclidean distance
    results = rnd.run_query(query, model)

    assert results["query_name"] == "Flowers and Insects wrt Pleasant"
    assert isinstance(results["result"], np.number)
    assert isinstance(results["rnd"], np.number)
    assert isinstance(results["distances_by_word"], dict)
    assert len(results["distances_by_word"]) > 0

    # test with cosine distance
    results = rnd.run_query(query, model, distance="cos")

    assert results["query_name"] == "Flowers and Insects wrt Pleasant"
    assert isinstance(results["result"], np.number)
    assert isinstance(results["rnd"], np.number)
    assert isinstance(results["distances_by_word"], dict)
    assert len(results["distances_by_word"]) > 0

    with pytest.raises(
        ValueError, match=r'distance_type can be either "norm" or "cos", .*'
    ):
        rnd.run_query(query, model, distance="other_distance")

    # lost word threshold test
    results = rnd.run_query(
        Query(
            [["bla", "asd"], weat_wordsets["insects"]],
            [weat_wordsets["pleasant_5"]],
            ["Flowers", "Insects"],
            ["Pleasant"],
        ),
        model,
    )
    assert results["query_name"] == "Flowers and Insects wrt Pleasant"
    assert np.isnan(results["result"])
    assert np.isnan(results["rnd"])
    assert isinstance(results["distances_by_word"], dict)
    assert len(results["distances_by_word"]) == 0


def test_RNSB(capsys, model, weat_wordsets):

    rnsb = RNSB()
    query = Query(
        [weat_wordsets["flowers"], weat_wordsets["insects"]],
        [weat_wordsets["pleasant_5"], weat_wordsets["unpleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant", "Unpleasant"],
    )
    results = rnsb.run_query(query, model)

    assert results["query_name"] == "Flowers and Insects wrt Pleasant and Unpleasant"
    assert list(results.keys()) == [
        "query_name",
        "result",
        "rnsb",
        "clf_accuracy",
        "negative_sentiment_probabilities",
        "negative_sentiment_distribution",
    ]
    assert isinstance(results["result"], (np.float32, np.float64, float, np.float_))
    assert isinstance(results["negative_sentiment_probabilities"], dict)
    assert isinstance(results["negative_sentiment_distribution"], dict)

    query = Query(
        [
            weat_wordsets["flowers"],
            weat_wordsets["instruments"],
            weat_wordsets["male_terms"],
            weat_wordsets["female_terms"],
        ],
        [weat_wordsets["pleasant_5"], weat_wordsets["unpleasant_5"]],
        ["Flowers", "Insects", "Male terms", "Female terms"],
        ["Pleasant", "Unpleasant"],
    )
    results = rnsb.run_query(query, model)

    assert (
        results["query_name"]
        == "Flowers, Insects, Male terms and Female terms wrt Pleasant and Unpleasant"
    )
    assert isinstance(results["result"], np.number)

    # custom classifier, print model eval
    results = rnsb.run_query(query, model, print_model_evaluation=True)

    print(capsys.readouterr())
    captured = capsys.readouterr()
    assert "Classification Report" in captured.out

    assert (
        results["query_name"]
        == "Flowers, Insects, Male terms and Female terms wrt Pleasant and Unpleasant"
    )
    assert isinstance(results["result"], np.number)
    assert isinstance(results["rnsb"], np.number)

    # lost word threshold test
    results = rnsb.run_query(
        Query(
            [["bla", "asd"], weat_wordsets["insects"]],
            [weat_wordsets["pleasant_5"], weat_wordsets["unpleasant_5"]],
            ["Flowers", "Insects"],
            ["Pleasant", "Unpleasant"],
        ),
        model,
    )
    assert np.isnan(results["rnsb"])
    assert np.isnan(results["result"])
    assert isinstance(results["negative_sentiment_probabilities"], dict)
    assert isinstance(results["negative_sentiment_distribution"], dict)
    assert len(results["negative_sentiment_probabilities"]) == 0
    assert len(results["negative_sentiment_distribution"]) == 0

    # test random state
    query = Query(
        [
            weat_wordsets["flowers"],
            weat_wordsets["instruments"],
            weat_wordsets["male_terms"],
            weat_wordsets["female_terms"],
        ],
        [weat_wordsets["pleasant_5"], weat_wordsets["unpleasant_5"]],
        ["Flowers", "Insects", "Male terms", "Female terms"],
        ["Pleasant", "Unpleasant"],
    )
    results = rnsb.run_query(query, model, random_state=42)

    assert (
        results["query_name"]
        == "Flowers, Insects, Male terms and Female terms wrt Pleasant and Unpleasant"
    )
    assert isinstance(results["result"], np.number)


def test_MAC(model, weat_wordsets):

    mac = MAC()
    query = Query(
        [weat_wordsets["flowers"]],
        [
            weat_wordsets["pleasant_5"],
            weat_wordsets["pleasant_9"],
            weat_wordsets["unpleasant_5"],
            weat_wordsets["unpleasant_9"],
        ],
        ["Flowers"],
        ["Pleasant 5 ", "Pleasant 9", "Unpleasant 5", "Unpleasant 9"],
    )
    results = mac.run_query(query, model)

    assert (
        results["query_name"]
        == "Flowers wrt Pleasant 5 , Pleasant 9, Unpleasant 5 and Unpleasant 9"
    )
    assert isinstance(results["result"], np.number)
    assert isinstance(results["mac"], np.number)
    assert isinstance(results["targets_eval"], dict)
    assert len(results["targets_eval"]["Flowers"]) == len(weat_wordsets["flowers"])
    # 4 = number of attribute sets
    assert len(results["targets_eval"]["Flowers"][weat_wordsets["flowers"][0]]) == 4

    # test metric with a target set that loses more words than allowed.
    query = Query(
        [weat_wordsets["flowers"], ["blabla", "asdf"]],
        [weat_wordsets["pleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant"],
    )
    results = mac.run_query(query, model)

    assert results["query_name"] == "Flowers and Insects wrt Pleasant"
    assert np.isnan(results["mac"])
    assert np.isnan(results["result"])


def test_ECT(model, weat_wordsets):

    ect = ECT()
    query = Query(
        [weat_wordsets["flowers"], weat_wordsets["insects"]],
        [weat_wordsets["pleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant"],
    )
    results = ect.run_query(query, model)

    assert results["query_name"] == "Flowers and Insects wrt Pleasant"
    assert isinstance(results["result"], np.number)
    assert isinstance(results["ect"], np.number)

    # test metric with a target set that loses more words than allowed.
    query = Query(
        [weat_wordsets["flowers"], ["blabla", "asdf"]],
        [weat_wordsets["pleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant"],
    )
    results = ect.run_query(query, model)

    assert results["query_name"] == "Flowers and Insects wrt Pleasant"
    assert np.isnan(results["ect"])
    assert np.isnan(results["result"])


def test_RIPA(model, weat_wordsets):

    ripa = RIPA()
    query = Query(
        [weat_wordsets["flowers"], weat_wordsets["insects"]],
        [weat_wordsets["pleasant_5"]],
        ["Flowers", "Insects"],
        ["Pleasant"],
    )
    results = ripa.run_query(query, model)

    assert results["query_name"] == "Flowers and Insects wrt Pleasant"
    assert isinstance(results["result"], (np.float32, np.float64, float))
    assert isinstance(results["ripa"], (np.float32, np.float64, float))
    assert isinstance(results["word_values"], dict)

    for word, word_value in results["word_values"].items():
        assert isinstance(word, str)
        assert isinstance(word_value, dict)
        assert isinstance(word_value["mean"], (np.float32, np.float64, float))
        assert isinstance(word_value["std"], (np.float32, np.float64, float))
